import json
import os.path
import sys

from bird_parser.parse_bird_output import parse_ospf_state_all

from jsonschema import validate


def main():
    if len(sys.argv) < 2:
        print(f"Usage: {sys.argv[0]} OSPF_LINK_DB_TXT_FILE [--no-validate]")
        sys.exit(2)

    file_name = sys.argv[1]
    if file_name == "-":
        input_str = sys.stdin.read()
    else:
        with open(file_name) as input_file:
            input_str = input_file.read()

    link_db = parse_ospf_state_all(input_str)

    if len(sys.argv) < 3 or sys.argv[2] != "--no-validate":
        schema_fname = os.path.join(os.path.dirname(__file__), "output_schema.json")
        with open(schema_fname, "r") as output_schema:
            validate(link_db, json.load(output_schema))

    print(json.dumps(link_db))


if __name__ == "__main__":
    main()



from typing import List, Any


class OutputDict(dict):
    def create_key_from_path(self, path: List[str]):
        nested_dict = self
        for item in path:
            if item not in nested_dict:
                nested_dict[item] = {}
            nested_dict = nested_dict[item]

    def set_value_by_path(self, path: List[str], value: Any):
        nested_dict = self
        for item in path[:-1]:
            nested_dict = nested_dict[item]

        nested_dict[path[-1]] = value

    def get_value_by_path(self, path: List[str]):
        nested_dict = self
        for item in path:
            nested_dict = nested_dict[item]

        return nested_dict


def parse_router_link(line_parts: List[str]):
    assert line_parts[2] in ["metric", "metric2"]
    link = {"id": line_parts[1], line_parts[2]: int(line_parts[3])}

    if len(line_parts) == 6:
        assert line_parts[4] == "via"
        link["via"] = line_parts[5]

    return link


def get_default_value_for_resource(resource_name: str):
    if resource_name == "area":
        return {
            "networks": {},
            "routers": {},
        }
    elif resource_name == "router":
        return {
            "links": {
                "network": [],
                "external": [],
                "router": [],
                "stubnet": [],
            }
        }
    elif resource_name == "network":
        return {
            "dr": "",
            "routers": [],
        }
    else:
        raise NotImplementedError(
            f"No default value found for resource: {resource_name}"
        )



import json
import os

import pytest
from bird_parser.parse_bird_output import parse_ospf_state_all


def test_with_live_data():
    absolute_path = os.path.dirname(__file__)
    input_fname = os.path.join(
        absolute_path, "artifacts/ospf_state_output_apr_29_2023.txt"
    )
    expected_output_fname = os.path.join(
        absolute_path, "artifacts/expected_output.json"
    )
    with open(input_fname, "r") as input_txt:
        parsed_object = parse_ospf_state_all(input_txt.read())

    with open(expected_output_fname, "r") as output_json:
        expected_output = json.load(output_json)

    parsed_routers = parsed_object["areas"]["0.0.0.0"]["routers"]
    expected_routers = expected_output["areas"]["0.0.0.0"]["routers"]

    for key in parsed_routers:
        assert parsed_routers[key] == expected_routers[key]

    parsed_networks = parsed_object["areas"]["0.0.0.0"]["networks"]
    expected_networks = expected_output["areas"]["0.0.0.0"]["networks"]

    for key in parsed_networks:
        assert parsed_networks[key] == expected_networks[key]



import re

from bird_parser.utils import (
    OutputDict,
    parse_router_link,
    get_default_value_for_resource,
)


def get_indentation_level(line: str):
    return len(line) - len(line.lstrip("\t"))


def parse_ospf_state_all(ospf_state_str: str):
    parser = StackParser(ospf_state_str)
    return parser.parse()


class StackParser:
    def __init__(self, ospf_state_str: str):
        self.output = OutputDict({})
        self.previous_indentation_level = 0
        self.path_stack = ["", ""]

        self.lines = [
            line.rstrip() for line in ospf_state_str.split("\n") if len(line) > 0
        ]
        self.lines.append("")

    def parse(self):
        assert re.match(r"BIRD.+ready\.", self.lines[0])

        for line in self.lines[1:]:
            self.parse_line(line)

        return self.output

    def parse_line(self, line: str):
        line_parts = line.split()
        indentation_level = get_indentation_level(line)

        if line_parts and (
            line_parts[0] == "distance" or line_parts[0] == "unreachable"
        ):
            # Ignore distance/reachability from our local perspective as it's not meaningful to
            # understanding the global state of the system
            return

        level_diff = indentation_level - self.previous_indentation_level
        if level_diff > 0:
            assert (
                level_diff == 1
            ), "Indenting by more than one tab at a time is not supported"
            self.handle_enter_resource(self.path_stack[-2][:-1])
        else:
            if level_diff < 0:
                self.handle_exit_resource(self.path_stack[2][:-1])

                # Account for the un-indent
                for i in range(abs(level_diff)):
                    self.path_stack.pop()
                    self.path_stack.pop()

            # Remove the previous "current level", since it will be replaced below
            self.path_stack.pop()
            self.path_stack.pop()

        if line_parts:
            self.path_stack.append(line_parts[0] + "s")
            self.path_stack.append(line_parts[1])

            current_container_name = (
                self.path_stack[-4][:-1] if len(self.path_stack) > 2 else None
            )
            if current_container_name == "router":
                link_type = line_parts[0]
                link_list_for_type = self.output.get_value_by_path(
                    self.path_stack[:-2] + ["links", link_type]
                )
                link_list_for_type.append(parse_router_link(line_parts))
            elif current_container_name == "network":
                network = self.output.get_value_by_path(self.path_stack[:-2])
                if line_parts[0] == "dr":
                    network["dr"] = line_parts[1]
                elif line_parts[0] == "router":
                    network["routers"].append(line_parts[1])
                else:
                    raise ValueError(f"Invalid line: {line}")

        self.previous_indentation_level = indentation_level

    def handle_enter_resource(self, resource_name: str):
        # Create an output key to track the new deeper indentation level
        self.output.create_key_from_path(self.path_stack)

        # If we are entering an area, router, or network, create the appropriate default containers
        if resource_name in ["area", "router", "network"]:
            self.output.set_value_by_path(
                self.path_stack,
                get_default_value_for_resource(resource_name),
            )

    def handle_exit_resource(self, resource_name: str):
        # If we are exiting a router, clean up the empty link lists so they don't appear in the output
        if resource_name == "router":
            link_lists = self.output.get_value_by_path(self.path_stack[:-2] + ["links"])
            lists_to_remove = [
                link_type for link_type, links in link_lists.items() if len(links) == 0
            ]
            for link_type in lists_to_remove:
                del link_lists[link_type]
